/*
    This program lists the available IL file formats.
    When invoked with '-heavy' it loads the dynamic shared object code
    for each format and prints information about the format.
*/

#include <stdio.h>
#include <stdlib.h>
#include <getopt.h>
#include <ifl/iflFormat.h>
#include <ifl/iflTypeNames.h>


void
printFormatInfo(iflDatabase* db, int longFormat)
{
    if (!longFormat) 
	printf("    %s (%s)%s\n", db->getDescription(), db->getName(), 
	       db->accessModeIsSupported(O_RDWR) ? "" : " (read only)");
    else {
	// get the full iflFormat so we can query everything
	iflFormat *fmt = db->getFormat();
	if (fmt == NULL) return;

	// print name and description
	printf("Format Name:\t%s\n", fmt->getName());
	printf("Description:\t%s\n", fmt->getDescription());
	
	// print supported access modes
	printf("Access modes:\t");
	if (fmt->accessModeIsSupported(O_RDWR))
	    printf("read/write\n");
	else if (fmt->accessModeIsSupported(O_RDONLY))
	    { printf("read only\n"); return; }
	else if (fmt->accessModeIsSupported(O_WRONLY))
	    printf("write only\n");
	else 
	    { printf("invalid\n"); return; }

	// print info on supported attributes
	printf("Supported attribute values:\n");
	
	printf("  Data types:\t");
	if (fmt->typeIsSupported(iflBit))
	    printf("bit ");
	if (fmt->typeIsSupported(iflChar))
	    printf("char ");
	if (fmt->typeIsSupported(iflUChar))
	    printf("uchar ");
	if (fmt->typeIsSupported(iflShort))
	    printf("short ");
	if (fmt->typeIsSupported(iflUShort))
	    printf("ushort ");
	if (fmt->typeIsSupported(iflInt))
	    printf("int ");
	if (fmt->typeIsSupported(iflUInt))
	    printf("uint ");
	if (fmt->typeIsSupported(iflFloat))
	    printf("float ");
	if (fmt->typeIsSupported(iflDouble))
	    printf("double");
	printf("\n");
	
	printf("  Orders:\t");
	if (fmt->orderIsSupported(iflInterleaved))
	    printf("interleaved ");
	if (fmt->orderIsSupported(iflSequential))
	    printf("sequential ");
	if (fmt->orderIsSupported(iflSeparate))
	    printf("separate");
	printf("\n");
	
	printf("  Orientations:\t");
	if (fmt->orientationIsSupported(iflUpperLeftOrigin))
	    printf("up-lft ");
	if (fmt->orientationIsSupported(iflLowerLeftOrigin))
	    printf("low-lft ");
	if (fmt->orientationIsSupported(iflUpperRightOrigin))
	    printf("up-rht ");
	if (fmt->orientationIsSupported(iflLowerRightOrigin))
	    printf("low-rht ");
	if (fmt->orientationIsSupported(iflLeftUpperOrigin))
	    printf("lft-up ");
	if (fmt->orientationIsSupported(iflLeftLowerOrigin))
	    printf("lft-low ");
	if (fmt->orientationIsSupported(iflRightUpperOrigin))
	    printf("rht-up ");
	if (fmt->orientationIsSupported(iflRightLowerOrigin))
	    printf("rht-low");
	printf("\n");

	printf("  Color models:\t");
	if (fmt->colorModelIsSupported(iflRGBPalette))
	    printf("palette ");
	if (fmt->colorModelIsSupported(iflLuminance))
	    printf("L ");
	if (fmt->colorModelIsSupported(iflLuminanceAlpha))
	    printf("LA ");
	if (fmt->colorModelIsSupported(iflNegative))
	    printf("negative ");
	if (fmt->colorModelIsSupported(iflRGB))
	    printf("RGB ");
	if (fmt->colorModelIsSupported(iflRGBA))
	    printf("RGBA ");
	if (fmt->colorModelIsSupported(iflBGR))
	    printf("BGR ");
	if (fmt->colorModelIsSupported(iflABGR))
	    printf("ABGR ");
	if (fmt->colorModelIsSupported(iflHSV))
	    printf("HSV ");
	if (fmt->colorModelIsSupported(iflCMY))
	    printf("CMY ");
	if (fmt->colorModelIsSupported(iflCMYK))
	    printf("CMYK ");
	if (fmt->colorModelIsSupported(iflYCC))
	    printf("YCC ");
	if (fmt->colorModelIsSupported(iflMultiSpectral))
	    printf("multi");
	printf("\n");

	printf("  Compressions:\t");
	if (fmt->compressionIsSupported(iflNoCompression))
	    printf("none ");
	if (fmt->compressionIsSupported(iflSGIRLE))
	    printf("SGI-RLE ");
	if (fmt->compressionIsSupported(iflCCITTFAX3))
	    printf("FAX3 ");
	if (fmt->compressionIsSupported(iflCCITTFAX4))
	    printf("FAX4 ");
	if (fmt->compressionIsSupported(iflLZW))
	    printf("LZW ");
	if (fmt->compressionIsSupported(iflJPEG))
	    printf("JPEG ");
	if (fmt->compressionIsSupported(iflPACKBITS))
	    printf("MAC-RLE ");
	if (fmt->compressionIsSupported(iflZIP))
	    printf("ZIP ");
	printf("\n");

	// print info on preferred attributes
	printf("Preferred attribute values:\n");
	
	printf("  Data type:\t%s\n", iflDataTypeName(fmt->getPreferredType()));
	printf("  Orientation:\t%s\n", 
	       iflOrientationName(fmt->getPreferredOrientation()));
	printf("  Order:\t%s\n", iflOrderName(fmt->getPreferredOrder()));
	printf("  Color model:\t%s\n", 
	       iflColorModelName(fmt->getPreferredColorModel()));
	printf("  Compression:\t%s\n", 
	       iflCompressionName(fmt->getPreferredCompression()));
    }
}


int
main(int argc, char **argv)
{
    // process command line arguments

    int usage = FALSE, longFormat = FALSE, c;
    
    while ((c = getopt(argc, argv, "l")) != -1) {
        switch (c) {
        case 'l':
            longFormat = TRUE;
            break;
        case '?':
            usage = TRUE;
            break;
	}
    }
	
    if (usage) {
        fprintf(stderr, "%s [-l] [<format-name-1> ...]\n", argv[0]);
        exit(0);
    }

    int numFormats = argc - optind;
    char** names = argv+optind;

    iflSetErrorHandler(iflRobustErrorHandler, NULL);

    // make sure we have some formats installed
    int index = 0;
    iflDatabase* db = iflDatabase::findNext(index);
    if (db == NULL)
	printf("No Image Format Library file format support installed.\n"
	       "Check the ifl_eoe release notes and the IFL(1) man page\n"
	       "for details on setting up the Image Format Library.\n");

    if (numFormats) {
	// print (long format) info on request formats

	for (int idx = 0; idx < numFormats; idx++) {
	    iflDatabase* db = iflDatabase::findByFormatName(names[idx]);
	    if (db == NULL) 
		printf("No format named %s installed, use %s with\n"
		       "no parameters to get a list of installed formats\n", 
		       names[idx], argv[0]);
	    else {
		if (idx > 0) printf("\n");
		printFormatInfo(db, TRUE);
	    }
	}
	
    }
    else {
	// print info on all installed formats

	printf("The following formats are supported by "
	       "the ImageVision Tools:\n");

	do {
	    if (longFormat) printf("\n");
	    printFormatInfo(db, longFormat);
	} while (db = iflDatabase::findNext(index));
    }

    return 0;
}
