/*
    Dump the image header info of any supported file format
*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <getopt.h>
#include <ifl/iflFile.h>
#include <ifl/iflFormat.h>
#include <ifl/iflTypeNames.h>


void 
printInfo(iflFile* file, int indexSpecified, int verbose)
{
    // print image name
    const char* name = file->getFileName();
    if (verbose)
        printf("Image file:\t\t%s\n", name);

    // get image dimensions
    iflSize dims;
    file->getDimensions(dims);

    // print file format
    iflFormat* format = file->getFormat();
    if (verbose)
	printf("File format:\t\t%s\n", format->getDescription());
    else 
	printf("%s:\t%s format (%s %dx%d)\n", name, format->getName(), 
	       iflColorModelName(file->getColorModel()), dims.x, dims.y);
    
    // the rest is only for long format
    if (!verbose) return;
	
    int nimg = file->getNumImgs();
    if (nimg > 1) printf("Number of sub-images:\t%d\n", nimg);

    // will print all images, unless an index was specified
    int firstimg = 0, lastimg = nimg-1, defaultimg = -1;
    if (indexSpecified)
	firstimg = lastimg = file->getCurrentImg();
    else
	defaultimg = file->getCurrentImg();
    
    for (int imgi = firstimg; imgi <= lastimg; imgi++) {
	file->setCurrentImg(imgi);	
	
	// print sub-image index (if more than 1)
	if (nimg > 1)
	    printf("\nSub-image:\t\t%d%s\n", imgi,
		   imgi == defaultimg ? "\t\t(default)" : "");

	// print image dimensions
	iflSize dims;
	file->getDimensions(dims);
	printf("Dimensions (w,h,z,c):\t"
	       "%d, %d, %d, %d\n", dims.x, dims.y, dims.z, dims.c);

	// print image pagesize
	iflSize pgDims;
	file->getPageDimensions(pgDims);
	printf("Page Size (w,h,z,c):\t"
	       "%d, %d, %d, %d\n", pgDims.x, pgDims.y, pgDims.z, pgDims.c);

	// print the rest of the image info
	printf("Data type:\t\t%s\n", iflDataTypeName(file->getDataType()));
	printf("Dimension order:\t%s\n", iflOrderName(file->getOrder()));
	printf("Color model:\t\t%s\n", 
	       iflColorModelName(file->getColorModel()));
	printf("Orientation:\t\t%s\n", 
	       iflOrientationName(file->getOrientation()));

	// print min and max values (if present)
	double min, max;
	if (file->getStatMinMax(min, max) == iflOKAY) 
	    printf("Statistical min/max:\t%g - %g\n", min, max);
	if (file->getScaleMinMax(min, max) == iflOKAY) 
	    printf("Display min/max:\t%g - %g\n", min, max);

	// print compression (if any)
	iflCompression compress = file->getCompression();
	if (compress != iflNoCompression) 
	    printf("Data Compression:\t%s\n", iflCompressionName(compress));

	// print description (if any)
	// char* msg = file->getDescription(TRUE);
	// if (msg) printf("Image description:\t%s\n", msg);
    }
}


void
main(int argc, char* argv[])
{
    // process command line arguments

    int usage = FALSE, verbose = FALSE, c;
    
    while ((c = getopt(argc, argv, "v")) != -1) {
        switch (c) {
        case 'v':
            verbose = TRUE;
            break;
        case '?':
            usage = TRUE;
            break;
	}
    }
	
    int numFiles = argc - optind;
    char** names = argv+optind;

    if (usage || numFiles < 1) {
        fprintf(stderr, "%s [-v] <image-file-1> [...]\n", argv[0]);
        exit(0);
    }

    // fore each image file, try to open the file and print info on it

    for (int idx = 0; idx < numFiles; idx++) {

	iflStatus sts;
	iflFile* file = iflFile::open(names[idx], O_RDONLY, &sts);
	if (file == NULL) {
	    char msg[1024];
	    iflStatusToString(sts, msg, sizeof msg);
	    fprintf(stderr, "Unable to open \"%s\": %s\n", names[idx], msg);
 	}
	else {
	    // leave a gap between files		
	    if (idx > 0 && verbose) printf("\n\n");
    
	    printInfo(file, strchr(names[idx], ':') != NULL, verbose ||
							     numFiles == 1);
    
	    if (file != NULL) file->close();
	}
    }
}
