#include <ifl/iflFile.h>
#include <ifl/iflFormat.h>
#include <ifl/iflError.h>
#include <ifl/iflDataSize.h>
#include <ifl/iflColor.h>
#include "iflFIT.h"
#if defined(_WIN32)
#   include <io.h>
#   define oserror() errno
    typedef __int64 off64_t;
#   define ssize_t int
#   define lseek64 _lseeki64
#else
#   include <unistd.h>
#endif
#include <fcntl.h>
#include <string.h>
#include <malloc.h>
#include <stdio.h>
#include <errno.h>

#if defined(__sgi)
#define FIT_BUILD_DIO
#define FIT_DIO_MIN_PAGESIZE	size_t(65536)
#endif

#if defined(__sgi) && IFL_HAVE_IRIX_RELEASE(6,5)
#define PREAD pread64
#define PWRITE pwrite64
#else
#define PREAD pread
#define PWRITE pwrite
#endif

#define ALIGN 128
#define roundup(p)	((p) + (ALIGN - (long(p) % ALIGN)))


class iflFITFile : public iflFile {

public:
    virtual iflStatus getItem(int tag, va_list ap);
    virtual iflStatus setItem(int tag, va_list ap);
    virtual iflStatus getScaleMinMax(double& min, double& max);
    virtual iflStatus setScaleMinMax(double min, double max);
    virtual iflStatus flush();

    virtual iflStatus getPage(void* data, int,int,int,int, int,int,int,int);
    virtual iflStatus setPage(const void* data, int,int,int,int,
			      int,int,int,int);

#if defined(__sgi) && defined(FIT_BUILD_DIO)
    virtual iflStatus _enableDirectIO(int enable);
    virtual iflStatus _getDirectIOMemAlignment(size_t& memAlign);
#endif

private:
    virtual iflStatus openFile();
    virtual iflStatus createFile();
    virtual iflStatus closeFile();
    friend class iflFITFormat;
    iflFITFile() {}  // only called by iflFITFormat::newfileobj()
    ~iflFITFile() {} // only called by iflFile::close()

    iflXYZint pageCount;	// number of pages in x,y,z dimensions
    size_t pageSizeBytes;	// number of bytes in each page
    int userOffset;		// byte offset to user extension
    int dataOffset;		// byte offset to first page
    int dataWritten;		// dataOffset can't be changed
    int needHeader;		// header not written flag

#if defined(FIT_BUILD_DIO)
    int dioEnab;		// direct I/O enabled flag
    int dioAvail;		// direct I/O available flag
#if defined(__sgi)
    int dioFd;			// direct I/O file descriptor
    struct dioattr dioAttr;	// direct I/O attributes for dioFd
#endif
#endif

#if !defined(__sgi)
    ssize_t pread(int fd, void* buf, size_t nbytes, off64_t offset)
	{
	    beginFileIO();
	    off64_t rc = lseek64(fd, offset, SEEK_SET);
	    if (rc != offset) return -1;
	    rc = read(fd, buf, nbytes);
	    endFileIO();
	    return (ssize_t)rc;
	}
    ssize_t pwrite(int fd, const void* buf, size_t nbytes, off64_t offset)
	{
	    beginFileIO();
	    off64_t rc = lseek64(fd, offset, SEEK_SET);
	    if (rc != offset) return -1;
	    rc = write(fd, buf, nbytes);
	    endFileIO();
	    return (ssize_t)rc;
	}
#endif

    double scaleMinValue;
    double scaleMaxValue;

    iflStatus writeHeader();
    iflStatus readHeader();

    void calcPageParams();

    int pageIndex(int x, int y, int z, int c) { 
	return ((c/pageSize.c*pageCount.z + z/pageSize.z)*pageCount.y + 
	        y/pageSize.y)*pageCount.x + x/pageSize.x; 
    }
    off64_t pageOffset(int x, int y, int z, int c)
	{return off64_t(pageIndex(x,y,z,c))*off64_t(pageSizeBytes)+dataOffset;}
};


class iflFITFormat : public iflFormat {
public:
    iflFITFormat() : iflFormat("FIT") {}
    int typeIsSupported(iflDataType) { return 1; }
    int orderIsSupported(iflOrder) { return 1; }
    int orientationIsSupported(iflOrientation) { return 1; }
    int colorModelIsSupported(iflColorModel) { return 1; }
    int compressionIsSupported(iflCompression cmp) 
	{ return cmp == iflNoCompression; }
    int sizeIsSupported(int,int,int,int, iflOrientation) 
	{ return 1;}
    int pagingIsSupported() { return 1; }
    int pageSizeIsSupported(int,int,int,int, int,int,int,int, iflOrientation) 
	{ return 1; }
    int randomAccessReadIsSupported() { return 1; }
    int randomAccessWriteIsSupported() { return 1; }
    virtual iflDataType getPreferredType() { return iflUChar; }
    virtual iflOrder getPreferredOrder() { return iflInterleaved; }
    iflOrientation getPreferredOrientation() { return iflLowerLeftOrigin; }
    virtual iflColorModel getPreferredColorModel(int nc) 
	{ return nc == 0? iflRGB : iflColorModelFromChans(nc); }
    virtual iflCompression getPreferredCompression() 
	{ return iflNoCompression; }
    void getPreferredPageSize(int, int, int, int sc,
			      int& pWidth, int& pHeight, int& pz, int& pc, 
			      iflOrientation)
	{ pWidth = 128; pHeight = 128; pz = 1; pc = sc; }
    iflFile* newfileobj() { return new iflFITFile; }
};


// This static object declaration causes the file format to be automatically
// registered when the DSO containing this code is dlopen'ed.
//
#if defined(__sgi)
#pragma set woff 1174	// 7.2 compiler doesn't understand /* REFERENCED */
#endif
/* REFERENCED */
static iflFITFormat theFITformat;
#if defined(__sgi)
#pragma reset woff 1174
#endif


const short iflFITmagic='IT';
const short iflFITversion='02';

struct FIThead02 {		// file header for version 02
    unsigned short magic;	// file ident
    unsigned short version;	// file version
    unsigned int xSize;		// image size
    unsigned int ySize;
    unsigned int zSize;
    unsigned int cSize;
    int dtype;			// data type
    int order;			// RGBRGB.. or RR..GG..BB..
    int space;			// coordinate space
    int cm;			// color model
    unsigned int xPageSize;	// page size
    unsigned int yPageSize;
    unsigned int zPageSize;
    unsigned int cPageSize;
				// NOTE: a word of padding is inserted here
				//       due to struct alignment rules
    double minValue;		// min/max pixel values
    double maxValue;
    unsigned int dataOffset;	// offset to first page of data
    // user extensible area...
};


struct FIThead01 {		// file header for version 01
    unsigned short magic;	// file ident
    unsigned short version;	// file version
    unsigned int xSize;		// image size
    unsigned int ySize;
    unsigned int zSize;
    unsigned int cSize;
    int dtype;			// data type
    int order;			// RGBRGB.. or RR..GG..BB..
    int space;			// coordinate space
    int cm;			// color model
    unsigned int xPageSize;	// page size
    unsigned int yPageSize;
    unsigned int zPageSize;
    unsigned int cPageSize;
    unsigned int dataOffset;	// offset to first page of data
    // user extensible area...
};


#if defined(_WIN32)
static inline void
SwabLong(unsigned int* lp)
{
    unsigned char* cp = (unsigned char*)lp;
    int t;
    t = cp[3]; cp[3] = cp[0]; cp[0] = t;
    t = cp[2]; cp[2] = cp[1]; cp[1] = t;
}


static inline void
SwabArrayOfShort(unsigned short* wp, unsigned int n)
{
    unsigned char* cp;
    int t;

    while (n-- > 0) {
	cp = (unsigned char*)wp;
	t = cp[1]; cp[1] = cp[0]; cp[0] = t;
	wp++;
    }
}


static inline void
SwabArrayOfShort(const unsigned short* wp, unsigned int n,
	unsigned short* dwp)
{
    const unsigned char* cp;
    unsigned char* dcp;

    while (n-- > 0) {
	cp = (const unsigned char*)wp;
	dcp = (unsigned char*)dwp;
	dcp[0] = cp[1];
	dcp[1] = cp[0];
	wp++; dwp++;
    }
}


static inline void
SwabArrayOfLong(unsigned int* lp, unsigned int n)
{
    unsigned char* cp;
    int t;
    while (n-- > 0) {
	cp = (unsigned char*)lp;
	t = cp[3]; cp[3] = cp[0]; cp[0] = t;
	t = cp[2]; cp[2] = cp[1]; cp[1] = t;
	lp++;
    }
}


static inline void
SwabArrayOfLong(const unsigned int* lp, unsigned int n, unsigned int* dlp)
{
    const unsigned char* cp;
    unsigned char* dcp;
    while (n-- > 0) {
	cp = (unsigned char*)lp;
	dcp = (unsigned char*)dlp;
	dcp[0] = cp[3];
	dcp[1] = cp[2];
	dcp[2] = cp[1];
	dcp[3] = cp[0];
	lp++; dlp++;
    }
}


static inline void
SwabArrayOfDouble(double* dp, unsigned int n)
{
    unsigned int* lp = (unsigned int*)dp;
    unsigned int t;

    SwabArrayOfLong(lp, n+n);
    while (n-- > 0) {
	t = lp[0]; lp[0] = lp[1]; lp[1] = t;
	lp += 2;
    }
}


static inline void
SwabArrayOfDouble(const double* dp, unsigned int n, double* ddp)
{
    const unsigned int* lp = (const unsigned int*)dp;
    unsigned int* dlp = (unsigned int*)ddp;
    unsigned int t;

    SwabArrayOfLong(lp, n+n, dlp);
    while (n-- > 0) {
	t = dlp[0]; dlp[0] = dlp[1]; dlp[1] = t;
	dlp += 2;
    }
}
#endif /* defined(_WIN32) */


iflStatus
iflFITFile::writeHeader()
{
    FIThead02 head;

    // clear the structure including padding, to get a deterministic result
    memset(&head, '\0', sizeof(head));

    // fill in the file header
    head.magic = iflFITmagic; 
    head.version = iflFITversion; 
    head.xSize = size.x; 
    head.ySize = size.y; 
    head.zSize = size.z; 
    head.cSize = size.c; 
    head.dtype = dtype;
    head.order = order; 
    head.space = orientation;
    head.cm = cm;
    head.xPageSize = pageSize.x;
    head.yPageSize = pageSize.y;
    head.zPageSize = pageSize.z;
    head.cPageSize = pageSize.c;
    head.minValue = scaleMinValue;
    head.maxValue = scaleMaxValue;
    head.dataOffset = dataOffset;

#if defined(_WIN32)
    SwabArrayOfShort(&head.magic, 2);
    SwabArrayOfLong(&head.xSize, 12);
    SwabArrayOfDouble(&head.minValue, 2);
    SwabLong(&head.dataOffset);
#endif

    lseek (fd, 0, SEEK_SET);
    if (write(fd, &head, sizeof head) == -1)
	return iflStatusEncode(iflWRITEFAILED, iflSubDomainUNIX, oserror());

    return iflOKAY;
}


iflStatus
iflFITFile::readHeader()
{
    FIThead02 head;

    lseek (fd, 0, SEEK_SET);
    if (read(fd, &head, sizeof head) == -1)
	return iflStatusEncode(iflREADFAILED, iflSubDomainUNIX, oserror());

#if defined(_WIN32)
    SwabArrayOfShort(&head.magic, 2);
#endif

    if (head.magic != iflFITmagic) return iflStatusEncode(iflBADMAGIC);

    // extract the image attributes from the file header
    if (head.version == '02') {
#if defined(_WIN32)
	SwabArrayOfLong(&head.xSize, 12);
	SwabArrayOfDouble(&head.minValue, 2);
	SwabLong(&head.dataOffset);
#endif
	size.x = head.xSize;
	size.y = head.ySize;
	size.z = head.zSize;
	size.c = head.cSize;
	dtype = iflDataType(head.dtype);	    // we can do this because
	order = iflOrder(head.order);		    // for some strange reason
	orientation = iflOrientation(head.space);   // FIT files use the exact
	cm = iflColorModel(head.cm);		    // same enum values as IFL
	pageSize.x = head.xPageSize;
	pageSize.y = head.yPageSize;
	pageSize.z = head.zPageSize;
	pageSize.c = head.cPageSize;
	scaleMinValue = head.minValue;
	scaleMaxValue = head.maxValue;
	dataOffset = head.dataOffset;
	userOffset = sizeof(FIThead02);
    }
    else if (head.version == '01') {
	FIThead01* head01 = (FIThead01*)&head;
#if defined(_WIN32)
	SwabArrayOfLong(&head01->xSize, 12);
	SwabLong(&head01->dataOffset);
#endif
	size.x = head01->xSize;
	size.y = head01->ySize;
	size.z = head01->zSize;
	size.c = head01->cSize;
	dtype = iflDataType(head01->dtype);
	order = iflOrder(head01->order);
	orientation = iflOrientation(head01->space);
	cm = iflColorModel(head01->cm+1);
	pageSize.x = head01->xPageSize;
	pageSize.y = head01->yPageSize;
	pageSize.z = head01->zPageSize;
	pageSize.c = head01->cPageSize;
	dataOffset = head01->dataOffset;
	userOffset = sizeof(FIThead01);
	scaleMinValue = iflDataMin(dtype);
	scaleMaxValue = iflDataMax(dtype);
    }
    else
	return iflStatusEncode(iflBADIMGFMT);

    return iflOKAY;
}


void 
iflFITFile::calcPageParams()
{
    pageCount.x = (size.x-1)/pageSize.x + 1;
    pageCount.y = (size.y-1)/pageSize.y + 1;
    pageCount.z = (size.z-1)/pageSize.z + 1;
    pageSizeBytes = iflDataSize(dtype, pageSize);
}


iflStatus iflFITFile::createFile()
{
    // validate the c page size
    if (pageSize.c == 0) 
	pageSize.c = order==iflSeparate? 1:size.c;
    else if (order != iflSeparate && pageSize.c != size.c) 
	return iflStatusEncode(iflBADPARAMS);

    if (fd < 0) {
	assert(filename != NULL);
#       if defined(_WIN32)
	    fd = ::open(filename, accessmode|O_CREAT|O_TRUNC|O_BINARY, 0666);
#       else
	    fd = ::open(filename, accessmode|O_CREAT|O_TRUNC, 0666);
#       endif
	if (fd < 0)
	    return iflStatusEncode(iflOPENFAILED, iflSubDomainUNIX, oserror());
    } else {
#	if !defined(_WIN32)
	    (void)::ftruncate(fd, (off_t)0);
#	endif
    }

    dataOffset = userOffset = sizeof(FIThead02);

    // flag the header/data as not written
    needHeader = 1;
    dataWritten = 0;

    calcPageParams();

    scaleMinValue = iflDataMin(dtype);
    scaleMaxValue = iflDataMax(dtype);

#if defined(FIT_BUILD_DIO)
#if defined(__sgi)
    if (pageSizeBytes >= FIT_DIO_MIN_PAGESIZE) {
	dioFd = (fcntl(fd, F_DIOINFO, &dioAttr) != -1) ? 
		::open(filename, accessmode|O_DIRECT) : -1;

	dioAvail = dioEnab = dioFd != -1;
    } else {
	dioEnab = dioAvail = FALSE;
	dioFd = -1;
    }
#else
    dioAvail = dioEnab = FALSE;
#endif
#endif

    return iflOKAY;
}


iflStatus iflFITFile::openFile()
{
    int fd_opened_here = 0;
    iflStatus status;

    if (fd < 0) {
	assert(filename != NULL);
#       if defined (_WIN32)
	    fd = (iflStatus)::open(filename, accessmode|O_BINARY);
#       else
	    fd = (iflStatus)::open(filename, accessmode);
#       endif
	if (fd < 0)
	    return iflStatusEncode(iflOPENFAILED, iflSubDomainUNIX, oserror());
	fd_opened_here = 1;
    }

    needHeader = 0;	// must be initialized for destructor!
    dataWritten = 1;	// so extensions can't be reserved

    // read the header
    if ((status = readHeader()) != iflOKAY) {
	if (fd_opened_here) {
	    (void)::close(fd);
	    fd = -1;
	}
	return status;
    }

    // fill in other info
    compression = iflNoCompression;
    calcPageParams();

#if defined(FIT_BUILD_DIO)
#if defined(__sgi)
    if (pageSizeBytes >= FIT_DIO_MIN_PAGESIZE) {
	dioFd = (fcntl(fd, F_DIOINFO, &dioAttr) != -1) ?
		::open(filename, accessmode|O_DIRECT) : -1;
	dioAvail = dioFd != -1;
	dioEnab = TRUE;
    } else {
	dioEnab = dioAvail = FALSE;
	dioFd = -1;
    }
#else
    dioAvail = dioEnab = FALSE;
#endif
#endif

    return iflOKAY;
}


// If this gets called, it's safe to
// assume that openFile() has been called and it succeeded.
//
iflStatus iflFITFile::closeFile()
{
    assert(fd >= 0);

#if defined(__sgi) && defined(FIT_BUILD_DIO)
    if (dioFd != -1) {
	::close(dioFd);
	dioFd = -1;
    }
#endif

    if (::close(fd) != 0)
	return iflStatusEncode(iflCLOSEFAILED, iflSubDomainUNIX, oserror());

    return iflOKAY;
}


iflStatus 
iflFITFile::getPage(void* data, int x, int y, int z, int c, int,int,int,int)
{
    iflStatus sts = iflOKAY;
    off64_t pgsoff = pageOffset(x,y,z,c);

#if defined(__sgi) && defined(FIT_BUILD_DIO)
    // Can't do direct IO if not enabled, buffer not cache aligned or
    // page not cache aligned
    if (!dioEnab || !dioAvail || (long)data % dioAttr.d_mem != 0 ||
            pgsoff % dioAttr.d_mem != 0 || pageSizeBytes < dioAttr.d_miniosz)
#endif
        sts = PREAD(fd, data, pageSizeBytes, pgsoff) != -1 ? iflOKAY :
                iflStatusEncode(iflREADFAILED, iflSubDomainUNIX, oserror());

#if defined(__sgi) && defined(FIT_BUILD_DIO)
    else {
	// Read any portion that is not block aligned
	off64_t pgeoff = pgsoff + pageSizeBytes;
	off_t slen = dioAttr.d_miniosz - (pgsoff % dioAttr.d_miniosz);
	off_t elen = pgeoff % dioAttr.d_miniosz;
	uchar_t* cdata = (uchar_t*)data;
        ssize_t rc;
	if (elen > 0) {
	    rc = PREAD(fd, cdata + (pgeoff - elen - pgsoff), elen,
		    pgeoff - elen);
	    if (rc == -1)
		return iflStatusEncode(iflREADFAILED, iflSubDomainUNIX,
			oserror());
	    pgeoff -= elen;
	}

	if (slen > 0) {
	    rc = PREAD(fd, data, slen, pgsoff);
	    if (rc == -1)
		return iflStatusEncode(iflREADFAILED, iflSubDomainUNIX,
			oserror());
	    pgsoff += slen;
	    cdata += slen;
	}

	off64_t len = pgeoff - pgsoff;
	if (len > 0) {
	    while (len > 0) {
		rc = PREAD(dioFd, cdata, len % dioAttr.d_maxiosz, pgsoff);
		if (rc == -1)
		    return iflStatusEncode(iflREADFAILED, iflSubDomainUNIX,
			    oserror());
		len -= dioAttr.d_maxiosz;
		pgsoff += dioAttr.d_maxiosz;
		cdata += dioAttr.d_maxiosz;
	    }
	}
    }
#endif
    
#if defined(_WIN32)
    switch (dtype) {
	case iflUShort:
	case iflShort:
	    SwabArrayOfShort((unsigned short*)data,
		    pageSizeBytes/sizeof(unsigned short));
	    break;
	case iflUInt:
	case iflInt:
	case iflFloat:
	    SwabArrayOfLong((unsigned int*)data,
		    pageSizeBytes/sizeof(unsigned int));
	    break;
	case iflDouble:
	    SwabArrayOfDouble((double*)data,
		    pageSizeBytes/sizeof(double));
	    break;
    }
#endif
    return sts;
}


iflStatus 
iflFITFile::setPage(const void* data, int x,int y,int z,int c,int,int,int,int)
{
    iflStatus sts = iflOKAY;
    off64_t pgsoff = pageOffset(x,y,z,c);

#if defined(_WIN32)
    unsigned char* tdata = NULL;
    if (dtype == iflUShort || dtype == iflShort || dtype == iflUInt ||
	    dtype == iflInt || dtype == iflFloat || dtype == iflDouble) {
	tdata = (unsigned char*)malloc(pageSizeBytes);
	switch (dtype) {
	    case iflUShort:
	    case iflShort:
		SwabArrayOfShort((unsigned short*)data,
			pageSizeBytes/sizeof(unsigned short),
			(unsigned short*)tdata);
		break;
	    case iflUInt:
	    case iflInt:
	    case iflFloat:
		SwabArrayOfLong((unsigned int*)data,
			pageSizeBytes/sizeof(unsigned int),
			(unsigned int*)tdata);
		break;
	    case iflDouble:
		SwabArrayOfDouble((double*)data,
			pageSizeBytes/sizeof(double),
			(double*)tdata);
		break;
	}
	data = tdata;
    }
#endif

#if defined(__sgi) && defined(FIT_BUILD_DIO)
    // Can't do direct IO if not enabled, buffer not cache aligned or
    // page not cache aligned
    if (!dioEnab || !dioAvail || (long)data % dioAttr.d_mem != 0 ||
            pgsoff % dioAttr.d_mem != 0 || pageSizeBytes < dioAttr.d_miniosz)
#endif
	sts = (dataWritten = PWRITE(fd, data, pageSizeBytes, pgsoff) != -1) ?
		iflOKAY :
		iflStatusEncode(iflWRITEFAILED, iflSubDomainUNIX, oserror());

#if defined(__sgi) && defined(FIT_BUILD_DIO)
    else {
	// Write any portion that is not block aligned
	off64_t pgeoff = pgsoff + pageSizeBytes;
	off_t slen = dioAttr.d_miniosz - (pgsoff % dioAttr.d_miniosz);
	off_t elen = pgeoff % dioAttr.d_miniosz;
	uchar_t* cdata = (uchar_t*)data;
        ssize_t rc;
	if (elen > 0) {
	    rc = PWRITE(fd, cdata + (pgeoff - elen - pgsoff), elen,
		    pgeoff - elen);
	    if (rc == -1)
		return iflStatusEncode(iflWRITEFAILED, iflSubDomainUNIX,
			oserror());
	    pgeoff -= elen;
	}

	if (slen > 0) {
	    rc = PWRITE(fd, data, slen, pgsoff);
	    if (rc == -1)
		return iflStatusEncode(iflWRITEFAILED, iflSubDomainUNIX,
			oserror());
	    dataWritten = 1;
	    pgsoff += slen;
	    cdata += slen;
	}

	off64_t len = pgeoff - pgsoff;
	if (len > 0) {
	    while (len > 0) {
		rc = PWRITE(dioFd, cdata, len % dioAttr.d_maxiosz, pgsoff);
		if (rc == -1)
		    return iflStatusEncode(iflWRITEFAILED, iflSubDomainUNIX,
			    oserror());
		dataWritten = 1;
		len -= dioAttr.d_maxiosz;
		pgsoff += dioAttr.d_maxiosz;
		cdata += dioAttr.d_maxiosz;
	    }
	}
    }
#endif

#if defined(_WIN32)
    if (tdata != NULL)
	free(tdata);
#endif

    return sts;
}


// In old IL code, there was a "discard" parameter.
// This is not needed here since there is no concept
// of in-memory page caching (that is assumed to be done
// at a higher level than IFL if at all).
//
iflStatus 
iflFITFile::flush()
{
    // update the header if necessary
    if (needHeader) { 
	iflStatus status = writeHeader(); 
	if (status != iflOKAY) return status;
	needHeader = 0;
    }
    return iflOKAY;
}


iflStatus 
iflFITFile::getScaleMinMax(double& min, double& max)
{
    min = scaleMinValue;
    max = scaleMaxValue;
    return iflOKAY;
}


iflStatus 
iflFITFile::setScaleMinMax(double min, double max)
{
    if (accessmode == O_RDONLY)
	return iflStatusEncode(iflREADONLY);
    scaleMinValue = min;
    scaleMaxValue = max;
    needHeader = 1;
    return iflOKAY;
}


iflStatus 
iflFITFile::getItem(int tag, va_list ap)
{
    switch (tag) {
    case iflFITextensionSize:
      {
	int* size = va_arg(ap, int*);
	if (size == NULL) return iflStatusEncode(iflBADPARAMS);
	*size = dataOffset-userOffset;
	return iflOKAY;
      }

    case iflFITextensionData: 
      {
	void* data = va_arg(ap, void*);
	int length = va_arg(ap, int);
	
	if (length > dataOffset-userOffset) 
	    return iflStatusEncode(iflBADPARAMS);
    
	lseek (fd, userOffset, SEEK_SET);
	if (read(fd, data, length) == -1)
	    return iflStatusEncode(iflREADFAILED, iflSubDomainUNIX, oserror());
    
	return iflOKAY;
      }

    default:
	return iflStatusEncode(iflBADPARAMS);
    }
}


iflStatus 
iflFITFile::setItem(int tag, va_list ap)
{
    switch (tag) {
    case iflFITextensionSize:
      {
	int size = va_arg(ap, int);

	// make room for user-specific data (if we still can)
	if (dataWritten) return iflStatusEncode(iflBADFIELDSET);
	dataOffset = userOffset + size;
    
	return iflOKAY;
      }

    case iflFITextensionData: 
      {
	void* data = va_arg(ap, void*);
	int length = va_arg(ap, int);
	
	// make room for user-specific data (if we still can)
	if (length > dataOffset-userOffset) {
	    if (dataWritten) return iflStatusEncode(iflBADPARAMS);
	    dataOffset = userOffset + length;
	}
    
	lseek (fd, userOffset, SEEK_SET);
	if (write(fd, data, length) == -1)
	    return iflStatusEncode(iflWRITEFAILED, iflSubDomainUNIX, oserror());
    
	return iflOKAY;
      }

    default:
	return iflStatusEncode(iflBADPARAMS);
    }
}

#if defined(FIT_BUILD_DIO)
iflStatus
iflFITFile::_enableDirectIO(int enable)
{
    if (dioAvail)
	dioEnab = enable != 0;
    return iflOKAY;
}


iflStatus
iflFITFile::_getDirectIOMemAlignment(size_t& memAlign)
{
#if defined(__sgi)
    memAlign = dioAvail ? dioAttr.d_mem : 0;
#else
    memAlign = 1;
#endif
    return iflOKAY;
}
#endif
