/*
    PNG i/o driver
*/
#include <stddef.h>
#include <stdlib.h>
#include <stdarg.h>
#include <errno.h>
#include <stdio.h>
#include <string.h>
#include <fcntl.h>
#include <setjmp.h>
#include <time.h>
#include <sys/types.h>
#if defined(_WIN32)
#include <io.h>
#define oserror() errno
#else

#ifndef SOLARIS_251
#include <inttypes.h>
#endif

#include <unistd.h>
#endif

#include <ifl/iflFormat.h>
#include <ifl/iflFile.h>
#include <ifl/iflColor.h>
#include <ifl/iflPixel.h>
#include <ifl/iflDataSize.h>
#include <ifl/iflTypeNames.h>
#include <ifl/iflError.h>

extern "C" {
#include "png.h"
}
#include "iflPNG.h"

class iflPNGFile : public iflFile {
public:
    virtual iflStatus flush();
    virtual iflStatus getPage(void* data, int,int,int,int, int,int,int,int);
    virtual iflStatus setPage(const void* data, int,int,int,int,
			      int,int,int,int);
    virtual iflStatus getScaleMinMax(double& min, double& max);

    virtual iflStatus getColormap(const iflColormap*&);
    virtual iflStatus setColormap(const iflColormap*);
    virtual iflStatus getItem(int tag, va_list);
    virtual iflStatus setItem(int tag, va_list);

private:
    virtual iflStatus openFile();
    virtual iflStatus createFile();
    virtual iflStatus closeFile();
    friend class iflPNGFormat;
    iflPNGFile() {}  // only called by iflPNGFormat::newfileobj()
    ~iflPNGFile() {} // only called by iflFile::close()

    static void errorHandler(png_struct *png, png_const_charp message);
    static void warningHandler(png_struct *png, png_const_charp message);

    FILE *fp;				// PNG file handle
    png_struct *png;			// PNG image handle
    png_info *info;			// PNG info handle
    iflColormap *cmap;			// colormap for palette images
    int npasses;			// number of interlace passes
    png_uint_16 *hist;			// palette histogram (write)
    png_byte *trans;			// palette transparency values (write)
};


class iflPNGFormat : public iflFormat {
public:
    iflPNGFormat() : iflFormat("PNG") {}
    int typeIsSupported(iflDataType type)
	{
	    return type == iflBit
		|| type == iflUChar
		|| type == iflUShort;
	}
    int orderIsSupported(iflOrder order)
	{ return order == iflInterleaved; }
    int orientationIsSupported(iflOrientation orientation)
	{ return orientation == iflUpperLeftOrigin; }
    int colorModelIsSupported(iflColorModel cm)
	{
	    return cm == iflLuminance
		|| cm == iflLuminanceAlpha
		|| cm == iflRGB
		|| cm == iflRGBPalette
		|| cm == iflRGBA;
	}
    int compressionIsSupported(iflCompression compress)
	{ return compress == iflZIP; }
    int sizeIsSupported(int /*width*/, int /*height*/, int z, int c,
			iflOrientation orientation)
	{
	    return z == 1
		&& c >= 1 && c <= 4
		&& orientation == iflUpperLeftOrigin;
	}
    int pagingIsSupported() { return FALSE; }
    int pageSizeIsSupported(int iWidth, int iHeight, int /*iz*/, int ic,
			    int pWidth, int pHeight, int pz, int pc,
			    iflOrientation /*orientation*/)
	{
	    return pWidth == iWidth && pHeight == iHeight
		&& pz == 1 && pc == ic;
	}
    int randomAccessReadIsSupported() { return FALSE; }
    int randomAccessWriteIsSupported() { return FALSE; }
    virtual iflDataType getPreferredType() { return iflUChar; }
    virtual iflOrder getPreferredOrder() { return iflInterleaved; }
    iflOrientation getPreferredOrientation() { return iflUpperLeftOrigin; }
    virtual iflColorModel getPreferredColorModel(int nc)
	{ return nc == 0? iflRGB : iflColorModelFromChans(nc); }
    virtual iflCompression getPreferredCompression() { return iflZIP; }
    void getPreferredPageSize(int iWidth, int iHeight, int, int iComp,
			      int& pWidth, int& pHeight, int& pz, int& pc, 
			      iflOrientation)
	{ pWidth = iWidth; pHeight = iHeight; pz = 1; pc = iComp; }
    iflFile* newfileobj() { return new iflPNGFile; }
    virtual iflTagDescription* _getTagDescriptions(int& count);
};


iflTagDescription*  
iflPNGFormat::_getTagDescriptions(int& count)
{
    static iflEnumSpec bitsSpec[] = {
    	{ "Match data type", 0 },
    	{ "Two", 2 },
    	{ "Four", 4 },
    	{ NULL, 0 }
    };
    static iflTagDescription desc[] = {
      { iflPNGbitspersample,
    	iflTagSet|iflTagEnum|iflTagOneParam|iflTagCreate,
    	"The number of bits stored per sample",
    	"Bits/sample",
    	"Match data type",
    	bitsSpec },
      { iflPNGbitspersample,
    	iflTagGet|iflTagInt|iflTagOneParam|iflTagInfo,
    	"The number of bits stored per sample",
    	"Bits/sample",
    	NULL,
    	NULL },
      { iflPNGinterlace,
    	iflTagGet|iflTagSet|iflTagBoolean|iflTagOneParam
    	|iflTagCreate|iflTagInfo,
    	"The image is stored in interlaced format",
    	"Interlace",
    	"false",
    	NULL }
    };

    count = sizeof desc / sizeof (iflTagDescription);
    return desc;
}


// This static object declaration causes the file format to be automatically
// registered when the DSO containing this code is dlopen'ed.
//
#if defined(__sgi)
#pragma set woff 1174	// 7.2 compiler doesn't understand /* REFERENCED */
#endif
/* REFERENCED */
static iflPNGFormat thePNGformat;
#if defined(__sgi)
#pragma reset woff 1174
#endif


/*static*/ void
iflPNGFile::errorHandler(png_struct *png, png_const_charp message)
{
    iflPNGFile *iflpng = (iflPNGFile *)png->error_ptr;
    iflError(MM_ERROR, "PNG file %s: %s", iflpng->filename, message);
    longjmp(png->jmpbuf, 1);
}


/*static*/ void
iflPNGFile::warningHandler(png_struct *png, png_const_charp message)
{
    iflPNGFile *iflpng = (iflPNGFile *)png->error_ptr;
    iflError(MM_WARNING, "PNG file %s: %s", iflpng->filename, message);
}


iflStatus
iflPNGFile::openFile()
{
    fp = NULL;	// so we can abort and call destructor at any time
    png = NULL;
    info = NULL;
    cmap = NULL;
    hist = NULL;
    trans = NULL;

    iflStatus status;

    // Open the file for reading.
    //
    if (accessmode != O_RDONLY) {
	// XXX This should be handle via accessModeIsSupported() but that
	// XXX function is inadequate to the task.
	iflError(MM_ERROR, "PNG file %s: access mode %d not supported for open",
		 filename, accessmode);
	status = iflStatusEncode(iflUNSUPPORTEDBYLIBRARY);
	goto error;
    }
    if (fd >= 0) {
	if (lseek(fd, (off_t)0, SEEK_SET) == (off_t)-1)
	    return iflStatusEncode(iflSEEKFAILED, iflSubDomainUNIX,oserror());
	fp = fdopen(fd, "rb");
    } else
	fp = fopen(filename, "rb");
    if (fp == NULL) {
	status = iflStatusEncode(iflOPENFAILED, iflSubDomainUNIX,oserror());
	goto error;
    }

    // Initialize PNG structures for reading the file.
    //
    png = png_create_read_struct(PNG_LIBPNG_VER_STRING, this, 
    	    	    	    	 errorHandler, warningHandler);
    if (png != NULL) info = png_create_info_struct(png);
    if (png == NULL || info == NULL) {
	status = iflStatusEncode(iflMALLOCFAILED, iflSubDomainUNIX,oserror());
	goto error;
    }

    if (setjmp(png->jmpbuf)) goto error;

    png_init_io(png, fp);
    png_read_info(png, info);

    // Have PNG library translate non-IFL image paradigms.
    //

    // Expand bit depths 2 and 4 to 8, and transparencies to full alpha
    // channels.  This scales the expanded values appropriately.  Note that
    // png_set_expand() also causes palette images to be expanded to RGB or
    // RGBA.
    if ((info->bit_depth > 1 && info->bit_depth < 8)
	|| info->valid & PNG_INFO_tRNS)
	png_set_expand(png);

    // Have PNG library handle progressive interlaced images.
    npasses = png_set_interlace_handling(png);

    // Update the info block with the changes brough about by doing the above
    // transformations.  Note this also calls png_start_read_image() so we
    // can't specify any more transformations.  Can we call this again if
    // we do need to add more PNG library transformations?
    png_read_update_info(png, info);

    // Extract the PNG image information into our iflFile instance.
    //
    dtype = info->bit_depth == 1 ? iflBit
	  : info->bit_depth == 8 ? iflUChar
	  : iflUShort;
    order = iflInterleaved;
    orientation = iflUpperLeftOrigin;
    switch (info->color_type) {
	case PNG_COLOR_TYPE_PALETTE:     cm = iflRGBPalette;      break;
	case PNG_COLOR_TYPE_RGB:         cm = iflRGB;             break;
	case PNG_COLOR_TYPE_GRAY:        cm = iflLuminance;       break;
	case PNG_COLOR_TYPE_RGB_ALPHA:   cm = iflRGBA;            break;
	case PNG_COLOR_TYPE_GRAY_ALPHA:  cm = iflLuminanceAlpha;  break;
	default:
	    iflError(MM_ERROR, "PNG file %s: unsupported PNG color model %d",
		     filename, info->color_type);
	    status = iflStatusEncode(iflUNSUPPORTEDBYLIBRARY);
	    goto error;
    }
    if (info->compression_type == 0)
	compression = iflZIP;
    else {
	iflError(MM_ERROR, "PNG file %s: unsupported PNG compression %d",
		 filename, info->compression_type);
	status = iflStatusEncode(iflUNSUPPORTEDBYLIBRARY);
	goto error;
    }
    size.x = pageSize.x = (int)info->width;
    size.y = pageSize.y = (int)info->height;
    size.z = pageSize.z = 1;
    size.c = pageSize.c = info->channels;
    if (size.c != iflColorModelChans(cm)) {
	iflError(MM_ERROR, "PNG file %s: wrong number of PNG channels (%d)"
		 " for color model %s",
		 filename, size.c, iflColorModelName(cm));
	status = iflStatusEncode(iflUNSUPPORTEDBYLIBRARY);
	goto error;
    }
    
    // for bit data the page size needs to be rounded up to a multiple of
    // eight to match PNG conventions
    if (dtype == iflBit) pageSize.x = (pageSize.x+7) & ~7;
    

    // Unpack image palette if image is color indexed.
    //
    if (cm == iflRGBPalette) {
	if ((info->valid & PNG_INFO_PLTE) == 0) {
	    iflError(MM_ERROR, "PNG file %s: missing PNG palette",
		     filename);
	    status = iflStatusEncode(iflBADIMGFMT);
	    goto error;
	}
	cmap = new iflColormap(3, iflUChar, 0, info->num_palette - 1);
	if (cmap == NULL) {
	    status = iflStatusEncode(iflMALLOCFAILED, iflSubDomainUNIX,oserror());
	    goto error;
	}
	unsigned char *rchan = (unsigned char *)cmap->getChan(0);
	unsigned char *gchan = (unsigned char *)cmap->getChan(1);
	unsigned char *bchan = (unsigned char *)cmap->getChan(2);
	png_byte *rgb = &info->palette->red;
	for (int cnt = info->num_palette; cnt > 0; cnt--) {
	    *rchan++ = *rgb++;
	    *gchan++ = *rgb++;
	    *bchan++ = *rgb++;
	}
    }

    return iflOKAY;

 error:
    png_destroy_read_struct(&png, &info, NULL);
    delete cmap;  cmap = NULL;

    // Clean up state for a failed open attempt.
    //
    if (fp != NULL) {
	if (fd > 0) {
	    // XXX The following bogosity is needed because we want to return
	    // XXX with fd still valid but there's no way to call fclose() on
	    // XXX fp without it also closing fd.  RFE 359885 has been filed
	    // XXX asking for this stdio capability.
	    fflush(fp);
	    fp->_file = ~0;
	}
	fclose(fp);
    }

    return status;
}


iflStatus
iflPNGFile::createFile()
{
    fp = NULL;	// so we can abort and call destructor at any time
    png = NULL;
    info = NULL;
    cmap = NULL;
    hist = NULL;
    trans = NULL;

    iflStatus status;

    // Open the file for writing.
    //
#if 0
    // XXX Need to fix accessModeIsSupported(); it's completely broken.
    // XXX We're getting O_RDWR here from IFL and there's no way to tell
    // XXX IFL that it's choice is wrong.
    if (accessmode != O_WRONLY) {
	// XXX This should be handle via accessModeIsSupported() but that
	// XXX function is inadequate to the task.
	iflError(MM_ERROR, "PNG file %s: access mode %d not supported for create",
		 filename, accessmode);
	status = iflStatusEncode(iflUNSUPPORTEDBYLIBRARY);
	goto error;
    }
#endif
    if (fd >= 0) {
	if (lseek(fd, (off_t)0, SEEK_SET) == (off_t)-1)
	    return iflStatusEncode(iflSEEKFAILED, iflSubDomainUNIX,oserror());
	fp = fdopen(fd, "wb");
    } else
	fp = fopen(filename, "wb");
    if (fp == NULL) {
	status = iflStatusEncode(iflOPENFAILED, iflSubDomainUNIX,oserror());
	goto error;
    }

    // Initialize PNG structures for writing the file.
    //
    png = png_create_write_struct(PNG_LIBPNG_VER_STRING, this, 
    	    	    	    	  errorHandler, warningHandler);
    if (png != NULL) info = png_create_info_struct(png);
    if (png == NULL || info == NULL) {
	status = iflStatusEncode(iflMALLOCFAILED, iflSubDomainUNIX,oserror());
	goto error;
    }

    if (setjmp(png->jmpbuf))  goto error;

    png_init_io(png, fp);

    // Copy image information from our iflFile instance into the PNG info
    // block.
    //
    info->bit_depth = dtype == iflBit ? 1 : dtype == iflUChar ? 8 : 16;
    switch (cm) {
	case iflRGBPalette:
	    info->color_type = PNG_COLOR_TYPE_PALETTE;
	    break;
	case iflRGB:
	    info->color_type = PNG_COLOR_TYPE_RGB;
	    break;
	case iflLuminance:
	    info->color_type = PNG_COLOR_TYPE_GRAY;
	    break;
	case iflRGBA:
            info->color_type = PNG_COLOR_TYPE_RGB_ALPHA;
	    break;
	case iflLuminanceAlpha:
	    info->color_type = PNG_COLOR_TYPE_GRAY_ALPHA;
	    break;
    }
    info->compression_type = 0;
    info->width = size.x;
    info->height = size.y;
    info->channels = size.c;

    // Delay calling png_write_info() till setPage() so user can call
    // setItem() to change various info fields.
    
    return iflOKAY;

 error:
    png_destroy_read_struct(&png, &info, NULL);

    // Clean up state for a failed open attempt.
    //
    if (fp != NULL) {
	if (fd > 0) {
	    // XXX The following bogosity is needed because we want to return
	    // XXX with fd still valid but there's no way to call fclose() on
	    // XXX fp without it also closing fd.  RFE 359885 has been filed
	    // XXX asking for this stdio capability.
	    fflush(fp);
	    fp->_file = ~0;
	}
	fclose(fp);
    }

    return status;
}


iflStatus
iflPNGFile::closeFile()
{
    if (png == NULL)
	return iflStatusEncode(iflOPENFAILED);

    if (png)
	if (accessmode == O_RDONLY)
	    png_read_destroy(png, info, (png_info *)NULL);
	else
	    png_write_destroy(png);
    if (fp) {
	fclose(fp);
	fp = NULL;
    }
    png_destroy_read_struct(&png, &info, NULL);
    delete cmap;   cmap = NULL;
    delete hist;   hist = NULL;
    delete trans;  trans = NULL;
    return iflOKAY;
}


iflStatus
iflPNGFile::flush()
{
    if (png == NULL)
	return iflStatusEncode(iflOPENFAILED);

    if (accessmode != O_RDONLY) {
	if (setjmp(png->jmpbuf)) 
    	    return iflStatusEncode(iflWRITEFAILED);
	png_write_flush(png);
    }
    
    return iflOKAY;
}


iflStatus 
iflPNGFile::getScaleMinMax(double& min, double& max)
{
    iflDataType scaleType;
    if (cm == iflRGBPalette && cmap != NULL)
    	scaleType = cmap->getDataType();
    else
    	scaleType = dtype;

    min = iflDataMin(scaleType);
    max = iflDataMax(scaleType);
    return iflOKAY;
}


iflStatus 
iflPNGFile::getPage(void* data, int x, int y, int z, int c,
		    int nx, int ny, int nz, int nc)
{
    if (png == NULL)
	return iflStatusEncode(iflOPENFAILED);

    if (x != 0 || y != 0 || z != 0 || c != 0
	|| nx != size.x || ny != size.y || nz != size.z || nc != size.c)
	return iflStatusEncode(iflUNSUPPORTEDBYFORMAT);

    if (setjmp(png->jmpbuf)) 
    	return iflOKAY;     // yeah, it failed, but let them have some 
    	    	    	    // data anyway, a messsage will be generated

    size_t rowsize = iflDataSize(dtype, size.x*size.c);
    for (int pass = 0; pass < npasses; pass++) {
	int y;
	png_byte *row;
	for (y = 0, row = (png_byte *)data; y < size.y; y++, row += rowsize)
	    png_read_row(png, row, NULL);
    }
    png_read_end(png, info);

    return iflOKAY;
}


iflStatus 
iflPNGFile::setPage(const void* data, int x, int y, int z, int c,
		    int nx, int ny, int nz, int nc)
{
    if (png == NULL)
	return iflStatusEncode(iflOPENFAILED);

    if (x != 0 || y != 0 || z != 0 || c != 0
	|| nx != size.x || ny != size.y || nz != size.z || nc != size.c)
	return iflStatusEncode(iflUNSUPPORTEDBYFORMAT);

    if (setjmp(png->jmpbuf)) 
    	return iflStatusEncode(iflWRITEFAILED);

    // Write the info header out.  Note that this may contain valid fields
    // other than those set in createFile() if the user called setItem()
    // with various ``interesting'' values.
    png_write_info(png, info);

    // Have PNG library translate non-IFL image paradigms.  Since
    // createFile() won't have set a non-IFL paradigm up, any such must
    // have been set up but the user calling setItem().
    //
    if (info->bit_depth > 1 && info->bit_depth < 8)
	png_set_packing(png);

    int npasses = png_set_interlace_handling(png);
    size_t rowsize = iflDataSize(dtype, size.x*size.c);
    for (int pass = 0; pass < npasses; pass++) {
	int y;
	png_byte *row;
	for (y = 0, row = (png_byte *)data; y < size.y; y++, row += rowsize)
	    png_write_row(png, row);
    }

    // By writing the PNG IEND chunk out here we lose the ability for users
    // to do a setPage() followed by a setItem(iflPNGtime, ...), etc.  I.e.
    // those special setItem()s must be done before the setPage().  If this
    // is a problem the png_write_end() call can be deferred till closeFile().
    //
    png_write_end(png, info);

    return iflOKAY;
}


iflStatus
iflPNGFile::getColormap(const iflColormap*& ret)
{
    if (png == NULL)
	return iflStatusEncode(iflOPENFAILED);

    ret = cmap;
    return iflOKAY;
}


iflStatus
iflPNGFile::setColormap(const iflColormap *newcmap)
{
    if (png == NULL)
	return iflStatusEncode(iflOPENFAILED);

    double min, max;
    if (accessmode == O_RDONLY || cm != iflRGBPalette || newcmap == NULL
	|| newcmap->getNumChans() != 3
	|| newcmap->getDomainMin() != 0 || newcmap->getDomainStep() != 1
	|| newcmap->getDomainMax() != newcmap->getLength()-1
	|| newcmap->getLength() > (1 << info->bit_depth)
	|| (newcmap->getRange(min, max), min < 0 || max > 255))
	return iflStatusEncode(iflBADPARAMS);

    if (cmap == NULL) cmap = new iflColormap;
    int length = newcmap->getLength();
    png_color *palette = new png_color[length];
    if (cmap == NULL || palette == NULL) {
	delete cmap;  cmap = NULL;
	delete palette;
	return iflStatusEncode(iflMALLOCFAILED, iflSubDomainUNIX,oserror());
    }

    *cmap = *newcmap;
    if (info->valid & PNG_INFO_PLTE)
	delete info->palette;
    info->palette = palette;
    info->num_palette = length;
    info->valid |= PNG_INFO_PLTE;

    png_byte *rgb = &info->palette->red;
    for (int i = 0; i < length; i++) {
	*rgb++ = (int)newcmap->getVal(i, 0);
	*rgb++ = (int)newcmap->getVal(i, 1);
	*rgb++ = (int)newcmap->getVal(i, 2);
    }

    return iflOKAY;
}


iflStatus
iflPNGFile::getItem(int tag, va_list ap)
{
    if (png == NULL)
	return iflStatusEncode(iflOPENFAILED);

    switch (tag) {
	default:
	    return iflStatusEncode(iflBADFIELDGET);
	case iflPNGhandles:
	    *va_arg(ap, png_struct **) = png;
	    *va_arg(ap, png_info **) = info;
	    break;
	case iflPNGbitspersample:
	    *va_arg(ap, int *) = info->bit_depth;
	    break;
	case iflPNGinterlace:
	    *va_arg(ap, int *) = info->interlace_type;
	    break;
	case iflPNGbackground: {
	    int *valid = va_arg(ap, int *);
	    iflPixel *background = va_arg(ap, iflPixel *);
	    if (background->getNumChans()
		!= (iflColorModelHasAlpha(cm) ? size.c - 1 : size.c)
		|| background->getDataType() != dtype)
		return iflStatusEncode(iflBADPARAMS);
	    *valid = (info->valid & PNG_INFO_bKGD) != 0;
	    if (*valid)
		switch (cm) {
		    case iflRGBPalette:
			background->setElem(info->background.index, 0);
			break;
		    case iflLuminance:
		    case iflLuminanceAlpha:
			background->setElem(info->background.gray, 0);
			break;
		    case iflRGB:
		    case iflRGBA:
			background->setElem(info->background.red, 0);
			background->setElem(info->background.green, 1);
			background->setElem(info->background.blue, 2);
			break;
		}
	    break;
	}
	case iflPNGchromaticity: {
	    int *valid = va_arg(ap, int *);
	    iflPNGchromaticity_t *chromaticity = va_arg(ap, iflPNGchromaticity_t *);
	    *valid = (info->valid & PNG_INFO_cHRM) != 0;
	    if (*valid) {
		chromaticity->xWhite = info->x_white;
		chromaticity->yWhite = info->y_white;
		chromaticity->xRed   = info->x_red;
		chromaticity->yRed   = info->y_red;
		chromaticity->xGreen = info->x_green;
		chromaticity->yGreen = info->y_green;
		chromaticity->xBlue  = info->x_blue;
		chromaticity->yBlue  = info->y_blue;
	    }
	    break;
	}
	case iflPNGgamma: {
	    int *valid = va_arg(ap, int *);
	    float *gamma = va_arg(ap, float *);
	    *valid = (info->valid & PNG_INFO_gAMA) != 0;
	    if (*valid)
		*gamma = info->gamma;
	    break;
	}
	case iflPNGhistogram: {
	    if (cm != iflRGBPalette)
		return iflStatusEncode(iflBADFIELDGET);
	    int *valid = va_arg(ap, int *);
	    png_uint_16 **histogram = va_arg(ap, png_uint_16 **);
	    *valid = (info->valid & PNG_INFO_hIST) != 0;
	    if (*valid)
		*histogram = info->hist;
	    break;
	}
	case iflPNGpixeldimensions: {
	    int *valid = va_arg(ap, int *);
	    iflPNGpixeldimensions_t *pdims = va_arg(ap, iflPNGpixeldimensions_t *);
	    *valid = (info->valid & PNG_INFO_pHYs) != 0;
	    if (*valid) {
		pdims->xPixelsPerUnit = (unsigned int)info->x_pixels_per_unit;
		pdims->yPixelsPerUnit = (unsigned int)info->y_pixels_per_unit;
		pdims->units = info->phys_unit_type;
	    }
	    break;
	}
	case iflPNGsignificantbits: {
	    int *valid = va_arg(ap, int *);
	    png_color_8 *sbits = va_arg(ap, png_color_8 *);
	    *valid = (info->valid & PNG_INFO_sBIT) != 0;
	    if (*valid)
		*sbits = info->sig_bit;
	    break;
	}
	case iflPNGtime: {
	    int *valid = va_arg(ap, int *);
	    struct tm *modtime = va_arg(ap, struct tm *);
	    *valid = (info->valid & PNG_INFO_tIME) != 0;
	    if (*valid) {
		modtime->tm_year  = info->mod_time.year;
		modtime->tm_mon   = info->mod_time.month;
		modtime->tm_mday  = info->mod_time.day;
		modtime->tm_hour  = info->mod_time.hour;
		modtime->tm_min   = info->mod_time.minute;
		modtime->tm_sec   = info->mod_time.second;
		modtime->tm_isdst = -1;
	    }
	    break;
	}
	case iflPNGtransparency: {
	    if (cm != iflRGB && cm != iflLuminance)
		return iflStatusEncode(iflBADFIELDGET);
	    int *valid = va_arg(ap, int *);
	    iflPixel *transparent = va_arg(ap, iflPixel *);
	    if (transparent->getNumChans() != (cm == iflRGB ? 3 : 1)
		|| transparent->getDataType() != dtype)
		return iflStatusEncode(iflBADPARAMS);
	    *valid = (info->valid & PNG_INFO_tRNS) != 0;
	    if (*valid) {
		if (cm == iflRGB) {
		    transparent->setElem(info->trans_values.red, 0);
		    transparent->setElem(info->trans_values.green, 1);
		    transparent->setElem(info->trans_values.blue, 2);
		} else
		    transparent->setElem(info->trans_values.gray, 0);
	    }
	    break;
	}
	case iflPNGpaletteTransparency: {
	    if (cm != iflRGBPalette)
		return iflStatusEncode(iflBADFIELDSET);
	    int *valid = va_arg(ap, int *);
	    int *nAlphas = va_arg(ap, int *);
	    png_byte **paletteAlphas = va_arg(ap, png_byte **);
	    *valid = (info->valid & PNG_INFO_tRNS) != 0;
	    if (*valid) {
		*nAlphas = info->num_trans;
		*paletteAlphas = info->trans;
	    }
	    break;
	}
    }
    return iflOKAY;
}


iflStatus
iflPNGFile::setItem(int tag, va_list ap)
{
    if (png == NULL)
	return iflStatusEncode(iflOPENFAILED);
    if (accessmode == O_RDONLY)
	return iflStatusEncode(iflREADONLY);

    switch (tag) {
	default:
	case iflPNGhandles: /* setItem(iflPNGhandles) not allowed */
	    return iflStatusEncode(iflBADFIELDSET);
	case iflPNGbitspersample: {
	    int bitspersample = va_arg(ap, int);
	    if (bitspersample == 0) return iflOKAY;
	    if (dtype != iflUChar || (bitspersample != 2 && bitspersample != 4))
		return iflStatusEncode(iflBADFIELDSET);
	    info->bit_depth = bitspersample;
	    break;
	}
	case iflPNGinterlace: {
	    int interlace = va_arg(ap, int);
	    if (interlace != 0 && interlace != 1)
		return iflStatusEncode(iflBADFIELDSET);
	    info->interlace_type = interlace;
	    break;
	}
	case iflPNGbackground: {
	    iflPixel *background = va_arg(ap, iflPixel *);
	    if (background->getNumChans()
		!= (iflColorModelHasAlpha(cm) ? size.c - 1 : size.c)
		|| background->getDataType() != dtype)
		return iflStatusEncode(iflBADPARAMS);
	    switch (cm) {
		case iflRGBPalette:
		    info->background.index = (int)background->getElem(0);
		    break;
		case iflLuminance:
		case iflLuminanceAlpha:
		    info->background.gray = (int)background->getElem(0);
		    break;
		case iflRGB:
		case iflRGBA:
		    info->background.red = (int)background->getElem(0);
		    info->background.green = (int)background->getElem(1);
		    info->background.blue = (int)background->getElem(2);
		    break;
	    }
	    info->valid |= PNG_INFO_bKGD;
	    break;
	}
	case iflPNGchromaticity: {
	    iflPNGchromaticity_t *chromaticity = va_arg(ap, iflPNGchromaticity_t *);
	    info->x_white = chromaticity->xWhite;
	    info->y_white = chromaticity->yWhite;
	    info->x_red   = chromaticity->xRed;
	    info->y_red   = chromaticity->yRed;
	    info->x_green = chromaticity->xGreen;
	    info->y_green = chromaticity->yGreen;
	    info->x_blue  = chromaticity->xBlue;
	    info->y_blue  = chromaticity->yBlue;
	    info->valid |= PNG_INFO_cHRM;
	    break;
	}
	case iflPNGgamma:
	    info->gamma = va_arg(ap, double);
	    info->valid |= PNG_INFO_gAMA;
	    break;
	case iflPNGhistogram: {
	    if (cm != iflRGBPalette)
		return iflStatusEncode(iflBADFIELDSET);
	    png_uint_16 *histogram = va_arg(ap, png_uint_16 *);
	    info->hist = histogram;
	    delete hist;  hist = histogram;
	    info->valid |= PNG_INFO_hIST;
	    break;
	}
	case iflPNGpixeldimensions: {
	    iflPNGpixeldimensions_t *pdims = va_arg(ap, iflPNGpixeldimensions_t *);
	    if (pdims->units != 0 && pdims->units != 1)
		return iflStatusEncode(iflBADPARAMS);
	    info->x_pixels_per_unit = pdims->xPixelsPerUnit;
	    info->y_pixels_per_unit = pdims->yPixelsPerUnit;
	    info->phys_unit_type = pdims->units;
	    info->valid |= PNG_INFO_pHYs;
	    break;
	}
	case iflPNGsignificantbits: {
	    png_color_8 *sbits = va_arg(ap, png_color_8 *);
	    int tooLarge = FALSE;
	    switch (cm) {
		case iflRGBA:
		    tooLarge = info->sig_bit.alpha >= info->bit_depth;
		    /*FALLTHROUGH*/
		case iflRGB:
		case iflRGBPalette:
		    tooLarge = tooLarge
			|| info->sig_bit.red >= info->bit_depth
			|| info->sig_bit.green >= info->bit_depth
			|| info->sig_bit.blue >= info->bit_depth;
		    break;
		case iflLuminanceAlpha:
		    tooLarge = info->sig_bit.alpha >= info->bit_depth;
		    /*FALLTHROUGH*/
		case iflLuminance:
		    tooLarge = tooLarge
			|| info->sig_bit.gray >= info->bit_depth;
		    break;
	    }
	    if (tooLarge)
		return iflStatusEncode(iflBADPARAMS);
	    info->sig_bit = *sbits;
	    info->valid |= PNG_INFO_sBIT;
	    break;
	}
	case iflPNGtime: {
	    struct tm *modtime = va_arg(ap, struct tm *);
	    info->mod_time.year   = modtime->tm_year;
	    info->mod_time.month  = modtime->tm_mon;
	    info->mod_time.day    = modtime->tm_mday;
	    info->mod_time.hour   = modtime->tm_hour;
	    info->mod_time.minute = modtime->tm_min;
	    info->mod_time.second = modtime->tm_sec;
	    info->valid |= PNG_INFO_tIME;
	    break;
	}
	case iflPNGtransparency: {
	    if (cm != iflRGB && cm != iflLuminance)
		return iflStatusEncode(iflBADFIELDSET);
	    iflPixel *transparent = va_arg(ap, iflPixel *);
	    if (transparent->getNumChans() != (cm == iflRGB ? 3 : 1)
		|| transparent->getDataType() != dtype)
		return iflStatusEncode(iflBADPARAMS);
	    if (cm == iflRGB) {
		info->trans_values.red   = (int)transparent->getElem(0);
		info->trans_values.green = (int)transparent->getElem(1);
		info->trans_values.blue  = (int)transparent->getElem(2);
	    } else
		info->trans_values.gray  = (int)transparent->getElem(0);
	    info->valid |= PNG_INFO_tRNS;
	    break;
	}
	case iflPNGpaletteTransparency: {
	    if (cm != iflRGBPalette)
		return iflStatusEncode(iflBADFIELDSET);
	    int nAlphas = va_arg(ap, int);
	    png_byte *paletteAlphas = va_arg(ap, png_byte *);
	    if (nAlphas > info->num_palette)
		return iflStatusEncode(iflBADPARAMS);
	    info->num_trans = nAlphas;
	    info->trans = paletteAlphas;
	    delete trans;  trans = paletteAlphas;
	    info->valid |= PNG_INFO_tRNS;
	    break;
	}
    }
    return iflOKAY;
}
